import aequitas
from aequitas.core.algorithms import AequitasTransformer

from aif360.algorithms.preprocessing import (
    DisparateImpactRemover as _Aif360DisparateImpactRemover,
    LFR as _Aif360LFR,
    Reweighing as _Aif360Reweighing,
    OptimPreproc as _Aif360OptimPreproc,
)


class DisparateImpactRemover(AequitasTransformer):
    def __init__(self, **kwargs):
        super().__init__(_Aif360DisparateImpactRemover, **kwargs)


class LFR(AequitasTransformer):
    def __init__(self, **kwargs):
        super().__init__(_Aif360LFR, **kwargs)


class Reweighing(AequitasTransformer):
    def __init__(self, **kwargs):
        super().__init__(_Aif360Reweighing, **kwargs)


class OptimPreproc(AequitasTransformer):
    def __init__(self, **kwargs):
        super().__init__(_Aif360OptimPreproc, **kwargs)


# keep this line at the bottom of this file
aequitas.logger.debug("Module %s correctly loaded", __name__)



import aequitas

from aequitas import Decorator
from aequitas.core.datasets import DatasetWithMetrics

from aif360.algorithms.transformer import Transformer

# from aif360.algorithms.inprocessing import (
#     AdversarialDebiasing,
#     ExponentiatedGradientReduction
# )
#
# from aif360.algorithms.preprocessing import (
#     DisparateImpactRemover,
#     LFR,
#     Reweighing
# )
#
# _ALGORITHM_TYPES = {
#     "adversarial debiasing": AdversarialDebiasing,
#     "egr": ExponentiatedGradientReduction,
#     "exponentiated gradient reduction": ExponentiatedGradientReduction,
#     "dir": DisparateImpactRemover,
#     "disparate impact remover": DisparateImpactRemover,
#     "lfr": LFR,
#     "reweighing": Reweighing
# }
#
#
# def create_algorithm(algorithm_type, **kwargs):
#     algorithm_type = algorithm_type.lower()
#     if algorithm_type not in _ALGORITHM_TYPES:
#         raise ValueError(f"Unknown algorithm type: {algorithm_type} \n Check among these: {_ALGORITHM_TYPES}")
#     return _ALGORITHM_TYPES[algorithm_type](**kwargs)


class AequitasTransformer(Decorator):
    def __init__(self, type, **kwargs):
        if not issubclass(type, Transformer):
            raise ValueError("Only AIF360 transformers can be decorated via this class")
        super().__init__(type(**kwargs))

    def _wrap_result(self, input: DatasetWithMetrics, result):
        if isinstance(result, DatasetWithMetrics):
            return result
        return input.wrap_another_dataset(result)

    def transform(self, dataset):
        return self._wrap_result(dataset, self._delegate.transform(dataset))

    def fit_transform(self, dataset):
        return self._wrap_result(dataset, self._delegate.fit_transform(dataset))


__all__ = [
    'AequitasTransformer',
]


# keep this line at the bottom of this file
aequitas.logger.debug("Module %s correctly loaded", __name__)
# keep this line at the bottom of this file
aequitas.logger.debug("Module %s correctly loaded", __name__)


import unittest
import numpy as np
import math

from aequitas.core.datasets import *
from aequitas.core.metrics import *

import tensorflow.compat.v1 as tf

tf.disable_eager_execution()


DEFAULT_TOLERANCE = 0.01


class AbstractMetricTestCase(unittest.TestCase):
    def assertInRange(self, value, lower, upper, msg=None, tolerance=DEFAULT_TOLERANCE):
        self.assertTrue(
            lower - tolerance <= value <= upper + tolerance,
            msg=msg or f"Value {value} should be in range [{lower}, {upper}]"
        )

    def assertNotInRange(self, value, lower, upper, msg=None, tolerance=DEFAULT_TOLERANCE):
        self.assertTrue(
            value < lower - tolerance or value > upper + tolerance,
            msg=msg or f"Value {value} should not be in range [{lower}, {upper}]"
        )

    def assertEDF(self, dataset, dataset_skewed, epsilon=2.337, tolerance=DEFAULT_TOLERANCE):
        score = dataset.metrics.smoothed_empirical_differential_fairness()
        self.assertFalse(math.isnan(score), msg="EDF should not be nan")
        self.assertIsNotNone(score, msg="EDF should not be None")
        self.assertInRange(score, lower=np.exp(-epsilon), upper=np.exp(epsilon), tolerance=tolerance)

        score = dataset_skewed.metrics.smoothed_empirical_differential_fairness()
        self.assertFalse(math.isnan(score), msg="EDF should not be nan")
        self.assertIsNotNone(score, msg="EDF should not be None")
        self.assertInRange(score, lower=np.exp(-epsilon), upper=np.exp(epsilon), tolerance=tolerance)

    def assertSP(self, dataset, dataset_skewed, bound=0.1, tolerance=DEFAULT_TOLERANCE):
        score = dataset_skewed.metrics.statistical_parity_difference()
        self.assertFalse(math.isnan(score), msg="SP should not be nan")
        self.assertIsNotNone(score, msg="SP should not be None")
        self.assertNotInRange(score, lower=-bound, upper=bound, tolerance=tolerance, msg="score should not be close to 0 for a biased dataset")

        score = dataset.metrics.statistical_parity_difference()
        self.assertFalse(math.isnan(score), msg="SP should not be nan")
        self.assertIsNotNone(score, msg="SP should not be None")
        self.assertInRange(score, -bound, bound, tolerance)

    def assertDI(self, dataset, dataset_skewed, bound=0.8, tolerance=DEFAULT_TOLERANCE):
        score = dataset.metrics.disparate_impact()
        self.assertFalse(math.isnan(score), msg="DI should not be nan")
        self.assertIsNotNone(score, msg="DI should not be None")
        self.assertGreaterEqual(score, bound - tolerance, f"DI should be >= {bound} for a nonbiased dataset")

        score = dataset_skewed.metrics.disparate_impact()
        self.assertFalse(math.isnan(score), msg="DI should not be nan")
        self.assertIsNotNone(score, msg="DI should not be None")
        self.assertLess(score, bound + tolerance, msg=f"DI is expected to be below {bound} for biased dataset")

    def assertConsistency(self, dataset, dataset_skewed, bound=0.8, tolerance=DEFAULT_TOLERANCE):
        score = dataset.metrics.disparate_impact()
        self.assertFalse(math.isnan(score), msg="Consistency should not be nan")
        self.assertIsNotNone(score, msg="Consistency should not be None")
        self.assertGreaterEqual(score, bound - tolerance, f"Consistency should be >= {bound} for a nonbiased dataset")

        score = dataset_skewed.metrics.disparate_impact()
        self.assertFalse(math.isnan(score), msg="Consistency should not be nan")
        self.assertIsNotNone(score, msg="Consistency should not be None")
        self.assertLess(score, bound + tolerance, msg=f"Consistency is expected to be < {bound} for biased dataset")

    def assertMeanDifference(self, dataset, dataset_skewed, bound=0.1, tolerance=DEFAULT_TOLERANCE):
        score = dataset.metrics.mean_difference()
        self.assertFalse(
            math.isnan(score),
            msg="Difference in mean outcomes between unprivileged and privileged groups should not be nan"
        )
        self.assertIsNotNone(
            score,
            msg="Difference in mean outcomes between unprivileged and privileged groups should not be None"
        )
        self.assertInRange(score, lower=-bound, upper=bound, tolerance=tolerance)

        score = dataset_skewed.metrics.mean_difference()
        self.assertFalse(
            math.isnan(score),
            msg="Difference in mean outcomes between unprivileged and privileged groups should not be nan"
        )
        self.assertIsNotNone(
            score,
            msg="Difference in mean outcomes between unprivileged and privileged groups should not be None"
        )
        self.assertLess(
            score,
            bound + tolerance,
            msg=f"Difference in mean outcomes between unprivileged and privileged should be < {-bound} in biased dataset"
        )

    def assertBinaryLabelDataset(self, dataset):
        self.assertIsInstance(dataset, DatasetWithMetrics)
        self.assertIsInstance(dataset._delegate, BinaryLabelDataset)
        self.assertIsInstance(dataset.metrics, BinaryLabelDatasetMetric)

    def assertMultiLabelDataset(self, dataset):
        self.assertIsInstance(dataset, DatasetWithMetrics)
        self.assertIsInstance(dataset._delegate, MulticlassLabelDataset)
        self.assertIsInstance(dataset.metrics, BinaryLabelDatasetMetric)

    def assertRegressionDataset(self, dataset):
        self.assertIsInstance(dataset, DatasetWithMetrics)
        self.assertIsInstance(dataset._delegate, RegressionDataset)
        self.assertIsInstance(dataset.metrics, RegressionDatasetMetric)



import unittest
import numpy as np
import pandas as pd

from aequitas.core.datasets import create_metric

from test import (
    generate_binary_label_dataframe,
    generate_skewed_binary_label_dataframe,
    generate_binary_label_dataframe_with_scores,
    generate_skewed_binary_label_dataframe_with_scores,
    generate_multi_label_dataframe,
    generate_skewed_multi_label_dataframe,
    generate_multi_label_dataframe_with_scores,
    generate_skewed_multi_label_dataframe_with_scores,
    generate_skewed_regression_dataset,
    generate_dataframe_with_preds
)
from test.core import AbstractMetricTestCase
from aif360.metrics import ClassificationMetric
from aequitas.core.imputation import *
from aequitas.core.datasets.zoo import *

import tensorflow.compat.v1 as tf

tf.disable_eager_execution()


class TestBinaryLabelDataset(AbstractMetricTestCase):
    def test_dataset_creation_via_factory(self):
        ds = create_dataset(
            "binary label",
            # parameters of aequitas.BinaryLabelDataset init
            unprivileged_groups=[{'prot_attr': 0}],
            privileged_groups=[{'prot_attr': 1}],
            imputation_strategy=MeanImputationStrategy(),
            df=generate_binary_label_dataframe()
        )
        self.assertBinaryLabelDataset(ds)

        ds_skewed = create_dataset(
            "binary label",
            # parameters of aequitas.BinaryLabelDataset init
            unprivileged_groups=[{'prot_attr': 0}],
            privileged_groups=[{'prot_attr': 1}],
            # parameters of aequitas.StructuredDataset init
            imputation_strategy=MeanImputationStrategy(),
            df=generate_skewed_binary_label_dataframe()
        )
        self.assertBinaryLabelDataset(ds_skewed)

    def test_dataset_creation_with_scores_via_factory(self):
        ds = create_dataset(
            "binary label",
            # parameters of aequitas.BinaryLabelDataset init
            unprivileged_groups=[{'prot_attr': 0}],
            privileged_groups=[{'prot_attr': 1}],
            # parameters of aequitas.StructuredDataset init
            imputation_strategy=MeanImputationStrategy(),
            # parameters of aif360.StructuredDataset init
            df=generate_binary_label_dataframe_with_scores(),
            scores_names="score"
        )
        self.assertBinaryLabelDataset(ds)

        ds_skewed = create_dataset(
            "binary label",
            # parameters of aequitas.BinaryLabelDataset init
            unprivileged_groups=[{'prot_attr': 0}],
            privileged_groups=[{'prot_attr': 1}],
            # parameters of aequitas.StructuredDataset init
            imputation_strategy=MeanImputationStrategy(),
            # parameters of aif360.StructuredDataset init
            df=generate_skewed_binary_label_dataframe_with_scores(),
            scores_names="score"
        )
        self.assertBinaryLabelDataset(ds_skewed)

    def test_metrics_on_dataset(self):
        df = generate_binary_label_dataframe_with_scores()
        df_skewed = generate_skewed_binary_label_dataframe_with_scores()
        ds = create_dataset(
            "binary label",
            # parameters of aequitas.BinaryLabelDataset init
            unprivileged_groups=[{'prot_attr': 0}],
            privileged_groups=[{'prot_attr': 1}],
            # parameters of aequitas.StructuredDataset init
            imputation_strategy=MeanImputationStrategy(),
            # parameters of aif360.StructuredDataset init
            df=df,
            scores_names="score"
        )
        self.assertBinaryLabelDataset(ds)

        ds_skewed = create_dataset(
            "binary label",
            # parameters of aequitas.BinaryLabelDataset init
            unprivileged_groups=[{'prot_attr': 0}],
            privileged_groups=[{'prot_attr': 1}],
            # parameters of aequitas.StructuredDataset init
            imputation_strategy=MeanImputationStrategy(),
            # parameters of aif360.StructuredDataset init
            df=df_skewed,
            scores_names="score"
        )
        self.assertBinaryLabelDataset(ds_skewed)

        ### METRICS USING LABELS ###

        # Disparate Impact
        self.assertDI(ds, ds_skewed)
        self.assertSP(ds, ds_skewed)
        # self.assertEDF(ds, ds_skewed)
        self.assertConsistency(ds, ds_skewed)

        ds_preds = create_dataset(
            "binary label",
            # parameters of aequitas.BinaryLabelDataset init
            unprivileged_groups=[{'prot_attr': 0}],
            privileged_groups=[{'prot_attr': 1}],
            imputation_strategy=MeanImputationStrategy(),
            df=generate_dataframe_with_preds(df),
            scores_names="score"
        )
        self.assertBinaryLabelDataset(ds_preds)

        cm = create_metric(
            metric="classification",
            dataset1=ds,
            dataset2=ds_preds,
            unprivileged_groups=[{'prot_attr': 0}],
            privileged_groups=[{'prot_attr': 1}]
        )

        self.assertIsInstance(cm, ClassificationMetric)


class TestMulticlassLabelDataset(AbstractMetricTestCase):
    def test_dataset_creation_via_factory(self):
        ds = create_dataset(
            "multi class",
            # parameters of aequitas.MulticlassLabelDataset init
            unprivileged_groups=[{'prot_attr': 0}],
            privileged_groups=[{'prot_attr': 1}],
            # parameters of aequitas.StructuredDataset init
            imputation_strategy=DoNothingImputationStrategy(),
            # parameters of aif360.MulticlassLabelDataset init
            favorable_label=[0, 1., 2.],
            unfavorable_label=[3., 4.],
            # parameters of aif360.StructuredDataset init
            df=generate_multi_label_dataframe()
        )
        self.assertMultiLabelDataset(ds)

        ds_skewed = create_dataset(
            "multi class",
            # parameters of aequitas.MulticlassLabelDataset init
            unprivileged_groups=[{'prot_attr': 0}],
            privileged_groups=[{'prot_attr': 1}],
            # parameters of aequitas.StructuredDataset init
            imputation_strategy=DoNothingImputationStrategy(),
            # parameters of aif360.MulticlassLabelDataset init
            favorable_label=[0, 1., 2.],
            unfavorable_label=[3., 4.],
            # parameters of aif360.StructuredDataset init
            df=generate_skewed_multi_label_dataframe()
        )
        self.assertMultiLabelDataset(ds)

        self.assertDI(ds, ds_skewed)
        self.assertSP(ds, ds_skewed)
        self.assertConsistency(ds, ds_skewed)

    def test_dataset_creation_with_scores_via_factory(self):
        ds = create_dataset(
            "multi class",
            # parameters of aequitas.MulticlassLabelDataset init
            unprivileged_groups=[{'prot_attr': 0}],
            privileged_groups=[{'prot_attr': 1}],
            # parameters of aequitas.StructuredDataset init
            imputation_strategy=DoNothingImputationStrategy(),
            # parameters of aif360.MulticlassLabelDataset init
            favorable_label=[0, 1., 2.],
            unfavorable_label=[3., 4.],
            # parameters of aif360.StructuredDataset init
            df=generate_multi_label_dataframe_with_scores(),
            scores_names="score"
        )
        self.assertMultiLabelDataset(ds)

        ds_skewed = create_dataset(
            "multi class",
            # parameters of aequitas.MulticlassLabelDataset init
            unprivileged_groups=[{'prot_attr': 0}],
            privileged_groups=[{'prot_attr': 1}],
            # parameters of aequitas.StructuredDataset init
            imputation_strategy=DoNothingImputationStrategy(),
            # parameters of aif360.MulticlassLabelDataset init
            favorable_label=[0, 1., 2.],
            unfavorable_label=[3., 4.],
            # parameters of aif360.StructuredDataset init
            df=generate_skewed_multi_label_dataframe_with_scores(),
            scores_names="score"
        )
        self.assertMultiLabelDataset(ds_skewed)


class TestRegressionDataset(AbstractMetricTestCase):
    def test_regression_dataset_creation_via_factory(self):
        ds = create_dataset(
            "regression",
            # parameters of aequitas.DatasetWithRegressionMetrics init
            unprivileged_groups=[{'color': 'b'}],
            privileged_groups=[{'color': 'r'}],
            # parameters of aequitas.StructuredDataset init
            imputation_strategy=MeanImputationStrategy(),
            # parameters of aif360.RegressionDataset init
            df=generate_skewed_regression_dataset(),
            dep_var_name='score'
        )
        self.assertRegressionDataset(ds)


class TestDataframeCreationFunctions(AbstractMetricTestCase):

    def setUp(self):
        self.nan_perc = 0.1

    def test_binary_dataframe_creation_with_nans(self):
        df = generate_binary_label_dataframe(nans=True)
        self.assertEqual(df.isna().sum().sum(), int(df.shape[0] * (df.shape[1] - 1) * self.nan_perc),
                         msg="nan counts do not correspond")

    def test_skewed_binary_dataframe_creation_with_nans(self):
        df = generate_skewed_binary_label_dataframe(nans=True)
        self.assertEqual(df.isna().sum().sum(), int(df.shape[0] * (df.shape[1] - 1) * self.nan_perc),
                         msg="nan counts do not correspond")

    def test_binary_dataframe_creation_with_nans_and_scores(self):
        df = generate_binary_label_dataframe_with_scores(nans=True)
        self.assertEqual(df.isna().sum().sum(), int(df.shape[0] * (df.shape[1] - 1) * self.nan_perc),
                         msg="nan counts do not correspond")

    def test_skewed_binary_dataframe_creation_with_nans_and_scores(self):
        df = generate_skewed_binary_label_dataframe_with_scores(nans=True)
        self.assertEqual(df.isna().sum().sum(), int(df.shape[0] * (df.shape[1] - 1) * self.nan_perc),
                         msg="nan counts do not correspond")

    def test_multi_label_dataframe_creation_with_nans(self):
        df = generate_multi_label_dataframe(nans=True)
        self.assertEqual(df.isna().sum().sum(), int(df.shape[0] * (df.shape[1] - 1) * self.nan_perc),
                         msg="nan counts do not correspond")

    def test_skewed_multi_label_dataframe_creation_with_nans(self):
        df = generate_skewed_multi_label_dataframe(nans=True)
        self.assertEqual(df.isna().sum().sum(), int(df.shape[0] * (df.shape[1] - 1) * self.nan_perc),
                         msg="nan counts do not correspond")

    def test_multi_label_dataframe_creation_with_nans_and_scores(self):
        df = generate_multi_label_dataframe(nans=True)
        self.assertEqual(df.isna().sum().sum(), int(df.shape[0] * (df.shape[1] - 1) * self.nan_perc),
                         msg="nan counts do not correspond")

    def test_skewed_multi_label_dataframe_creation_with_nans_and_scores(self):
        df = generate_skewed_multi_label_dataframe(nans=True)
        self.assertEqual(df.isna().sum().sum(), int(df.shape[0] * (df.shape[1] - 1) * self.nan_perc),
                         msg="nan counts do not correspond")


if __name__ == '__main__':
    unittest.main()



import unittest
import numpy as np
import pandas as pd

from sklearn.preprocessing import MinMaxScaler, MaxAbsScaler
from sklearn.linear_model import LogisticRegression

from test.core import AbstractMetricTestCase
from aequitas.core.imputation import *
from aequitas.core.algorithms.preprocessing import *
from aequitas.core.algorithms.postprocessing import *
from aequitas.core.algorithms.inprocessing import *
from aequitas.core.datasets.zoo import *

import tensorflow.compat.v1 as tf

tf.disable_eager_execution()


class TestMitigationAlgorithms(AbstractMetricTestCase):

    def test_disparate_impact_remover_on_adult_dataset(self):
        protected = "sex"

        ds = adult(unprivileged_groups=[{protected: 0}],
                   privileged_groups=[{protected: 1}],
                   protected_attribute_names=[protected],
                   privileged_classes=[['Male']],
                   categorical_features=[],
                   features_to_keep=['age', 'education-num', 'capital-gain', 'capital-loss', 'hours-per-week'])

        scaler = MinMaxScaler(copy=False)

        test, train = ds.split([16281])

        train.features = scaler.fit_transform(train.features)
        test.features = scaler.fit_transform(test.features)

        index = train.feature_names.index(protected)

        # create_algorithm(algorithm_type="dir", sensitive_attribute=protected)
        di_remover = DisparateImpactRemover(sensitive_attribute=protected)

        train_repd = di_remover.fit_transform(train)
        test_repd = di_remover.fit_transform(test)

        X_tr = np.delete(train_repd.features, index, axis=1)
        X_te = np.delete(test_repd.features, index, axis=1)
        y_tr = train_repd.labels.ravel()

        lmod = LogisticRegression(class_weight='balanced', solver='liblinear')
        lmod.fit(X_tr, y_tr)

        test_repd_pred = test_repd.copy()
        test_repd_pred.labels = lmod.predict(X_te)

        self.assertDI(dataset=test_repd_pred, dataset_skewed=test)

    def test_reweighing_on_adult_dataset(self):
        ds = adult()
        rw = Reweighing(unprivileged_groups=ds.unprivileged_groups, privileged_groups=ds.privileged_groups)
        repaired_ds = rw.fit_transform(ds)
        self.assertMeanDifference(dataset=repaired_ds, dataset_skewed=ds)

    def test_adversarial_debiasing_on_adult_dataset(self):
        ds = adult()
        train, test = ds.split([0.7], shuffle=True)

        min_max_scaler = MaxAbsScaler()
        train.features = min_max_scaler.fit_transform(train.features)
        test.features = min_max_scaler.transform(test.features)

        sess = tf.Session()
        plain_model = AdversarialDebiasing(
            privileged_groups=ds.privileged_groups,
            unprivileged_groups=ds.unprivileged_groups,
            scope_name='plain_classifier',
            debias=False,
            sess=sess
        )

        plain_model.fit(train)

        nondebiasing_train = plain_model.predict(train)
        nondebiasing_test = plain_model.predict(test)

        sess.close()
        tf.reset_default_graph()
        sess = tf.Session()

        debiased_model = AdversarialDebiasing(
            privileged_groups=ds.privileged_groups,
            unprivileged_groups=ds.unprivileged_groups,
            scope_name='debiased_classifier',
            debias=True,
            sess=sess
        )
        debiased_model.fit(train)

        debiasing_train = debiased_model.predict(train)
        debiasing_test = debiased_model.predict(test)

        self.assertMeanDifference(dataset=debiasing_train, dataset_skewed=nondebiasing_train)
        self.assertMeanDifference(dataset=debiasing_test, dataset_skewed=nondebiasing_test)

    def test_deterministic_reranking(self):
        balls = pd.DataFrame(
            [['r', 100], ['r', 90], ['r', 85], ['r', 70], ['b', 70], ['b', 60], ['b', 50], ['b', 40], ['b', 30],
             ['r', 20]],
            columns=['color', 'score'])

        balls_ds = create_dataset(
            dataset_type="regression",
            unprivileged_groups=[{'color': 'b'}],
            privileged_groups=[{'color': 'r'}],
            # parameters of aequitas.StructuredDataset init
            imputation_strategy=DoNothingImputationStrategy(),
            # parameters of aif360.RegressionDataset init
            df=balls,
            dep_var_name='score')

        balls_ds.labels = np.transpose([balls['score']])

        dr = DeterministicReranking(unprivileged_groups=[{'color': 0}], privileged_groups=[{'color': 1}])

        fair_ranking = dr.fit_predict(dataset=balls_ds, rec_size=6, target_prop=[0.5, 0.5], rerank_type='Constrained')
        fair_ranking.convert_to_dataframe()[0]


if __name__ == '__main__':
    unittest.main()

# TODO: test ClassificationMetric, test RegressionMetric



import aequitas
from aequitas.core.algorithms import AequitasTransformer

from aif360.algorithms.inprocessing import (
    AdversarialDebiasing as _Aif360AdversarialDebiasing,
    ExponentiatedGradientReduction as _Aif360ExponentiatedGradientReduction,
    GridSearchReduction as _Aif360GridSearchReduction,
    PrejudiceRemover as _Aif360PrejudiceRemover,
    ARTClassifier as _Aif360ARTClassifier,
    GerryFairClassifier as _Aif360GerryFairClassifier,
    MetaFairClassifier as _Aif360MetaFairClassifier,
)


class AdversarialDebiasing(AequitasTransformer):
    def __init__(self, **kwargs):
        super().__init__(_Aif360AdversarialDebiasing, **kwargs)


class ExponentiatedGradientReduction(AequitasTransformer):
    def __init__(self, **kwargs):
        super().__init__(_Aif360ExponentiatedGradientReduction, **kwargs)


class GridSearchReduction(AequitasTransformer):
    def __init__(self, **kwargs):
        super().__init__(_Aif360GridSearchReduction, **kwargs)


class PrejudiceRemover(AequitasTransformer):
    def __init__(self, **kwargs):
        super().__init__(_Aif360PrejudiceRemover, **kwargs)


class ARTClassifier(AequitasTransformer):
    def __init__(self, **kwargs):
        super().__init__(_Aif360ARTClassifier, **kwargs)


class GerryFairClassifier(AequitasTransformer):
    def __init__(self, **kwargs):
        super().__init__(_Aif360GerryFairClassifier, **kwargs)


class MetaFairClassifier(AequitasTransformer):
    def __init__(self, **kwargs):
        super().__init__(_Aif360MetaFairClassifier, **kwargs)


# keep this line at the bottom of this file
aequitas.logger.debug("Module %s correctly loaded", __name__)



import aequitas
from aequitas.core.datasets import create_dataset, StandardDataset, DatasetWithBinaryLabelMetrics, DatasetWithRegressionMetrics
from aif360.algorithms.preprocessing.optim_preproc_helpers.data_preproc_functions import load_preproc_data_adult
from pathlib import Path
import urllib.request
import aif360

DATA_DIR = Path(aif360.__file__).parent / "data" / "raw"


def _download(url: str, path: Path, timeout=60*5, buffer_size=1024*1024):
    aequitas.logger.warning(f"Downloading {url} to {path}")
    with open(path, 'wb') as file:
        with urllib.request.urlopen(url, timeout=timeout) as response:
            if 200 <= response.getcode() < 300:
                while True:
                    buffer = response.read(buffer_size)
                    if not buffer:
                        aequitas.logger.info(f"Successfully downloaded {url} to {path}")
                        return
                    file.write(buffer)
            raise Exception(f"Failed to download {url}: {response.getcode()}")


def _download_if_missing(dirname: str, filename: str, baseurl: str, max_attempts: int = 3):
    dir = DATA_DIR / dirname
    if not dir.exists():
        dir.mkdir()
    file = dir / filename
    if not file.exists():
        url = f"{baseurl}/{filename}"
        while max_attempts > 0:
            try:
                _download(url, file)
                return
            except Exception as e:
                max_attempts -= 1
                if max_attempts == 0:
                    raise e
                else:
                    aequitas.logger.warning(
                        f"Failed to download {url} in {file}. Retrying ({max_attempts} remaining attempts). Cause: {e}"
                    )


def adult(unprivileged_groups=None, privileged_groups=None, **kwargs) -> DatasetWithBinaryLabelMetrics:
    unprivileged_groups = unprivileged_groups or [{'sex': 0}]
    privileged_groups = privileged_groups or [{'sex': 1}]
    for name in ("adult.data", "adult.test", "adult.names"):
        _download_if_missing("adult", name, "https://archive.ics.uci.edu/ml/machine-learning-databases/adult")
    protected_attributes = kwargs.pop('protected_attributes', ['sex'])
    sub_samp = kwargs.pop('sub_samp', None)
    balance = kwargs.pop('balance', None)
    dataset: StandardDataset = load_preproc_data_adult(protected_attributes, sub_samp, balance)
    return create_dataset("binary",
                          unprivileged_groups=unprivileged_groups,
                          privileged_groups=privileged_groups,
                          wrap=dataset,
                          **kwargs)



from aequitas.core.datasets import DatasetWithBinaryLabelMetrics, BinaryLabelDataset
from aequitas.core.datasets.zoo import adult
from aequitas.core.metrics import BinaryLabelDatasetMetric
import unittest


class TestBinaryLabelDataset(unittest.TestCase):
    def setUp(self) -> None:
        protected = 'sex'
        self._dataset = adult(
            unprivileged_groups=[{protected: 0}],
            privileged_groups=[{protected: 1}],
            protected_attribute_names=[protected],
            privileged_classes=[['Male']],
        )

    def test_types(self):
        self.assertIsInstance(self._dataset, DatasetWithBinaryLabelMetrics)
        self.assertIsInstance(self._dataset._delegate, BinaryLabelDataset)

    def test_metrics(self):
        self.assertIsInstance(self._dataset.metrics, BinaryLabelDatasetMetric)

    def test_copy_preserves_type(self):
        copied = self._dataset.copy()
        self.assertIsInstance(copied, DatasetWithBinaryLabelMetrics)
        self.assertIsInstance(copied._delegate, BinaryLabelDataset)
        self.assertIsInstance(copied.metrics, BinaryLabelDatasetMetric)

    def test_split_preserves_type(self):
        splitted = self._dataset.split([0.5])
        for split in splitted:
            self.assertIsInstance(split, DatasetWithBinaryLabelMetrics)
            self.assertIsInstance(split._delegate, BinaryLabelDataset)
            self.assertIsInstance(split.metrics, BinaryLabelDatasetMetric)
    


from abc import ABC, abstractmethod
import aequitas
import pandas
import pandas as pd


class MissingValuesImputationStrategy(ABC):
    def __call__(self, df: pandas.DataFrame) -> pd.DataFrame:
        raise NotImplementedError("Implement this method in a subclass")


class MeanImputationStrategy(MissingValuesImputationStrategy):
    def __call__(self, df: pandas.DataFrame) -> pd.DataFrame:
        return df.fillna(df.mean(numeric_only=True))
    

class DoNothingImputationStrategy(MissingValuesImputationStrategy):
    def __call__(self, df: pandas.DataFrame) -> pd.DataFrame:
        return df


class MCMCImputationStrategy(MissingValuesImputationStrategy):
    def __call__(self, df: pandas.DataFrame) -> pd.DataFrame:
        import warnings
        warnings.warn("Unimplemented MCMC imputation strategy: doing nothig")
        return df


# keep this line at the bottom of this file
aequitas.logger.debug("Module %s correctly loaded", __name__)


# noinspection PyUnresolvedReferences
from aif360.datasets import (
    Dataset,
    StructuredDataset,
    RegressionDataset,
    BinaryLabelDataset,
    StandardDataset
)

# noinspection PyUnresolvedReferences
from aif360.datasets.multiclass_label_dataset import MulticlassLabelDataset

import aequitas
import aequitas.core.metrics as _metrics
import aequitas.core.imputation as _imputations


class DatasetWithMetrics(aequitas.Decorator):
    def __init__(self, decorated: StructuredDataset, metrics_type: type):
        if not isinstance(decorated, StructuredDataset):
            raise ValueError("Only structured datasets can be decorated via this class")
        super().__init__(decorated)
        self._metrics_type = metrics_type

    def _new_metrics(self) -> _metrics.Metric:
        return self._metrics_type(dataset=self._delegate)

    @property
    def metrics(self) -> _metrics.Metric:
        return self._new_metrics()

    @property
    def delegate(self) -> aequitas.Decorator:
        return self._delegate

    def _attributes_to_propagate_when_wrapping(self):
        return []

    def wrap_another_dataset(self, dataset: StructuredDataset, **kwargs) -> 'DatasetWithMetrics':
        propagated = {k: getattr(self, k) for k in self._attributes_to_propagate_when_wrapping()}
        kwargs.update(propagated)
        return self.__class__(dataset, **kwargs)

    def split(self, *args, **kwargs):
        return tuple(
            self.wrap_another_dataset(splitted)
            for splitted in self._delegate.split(*args, **kwargs)
        )

    def copy(self, *args, **kwargs):
        return self.wrap_another_dataset(self._delegate.copy(*args, **kwargs))


class DatasetWithBinaryLabelMetrics(DatasetWithMetrics):
    def __init__(self, dataset, unprivileged_groups, privileged_groups):
        super().__init__(dataset, _metrics.BinaryLabelDatasetMetric)
        self.unprivileged_groups = unprivileged_groups
        self.privileged_groups = privileged_groups

    def _new_metrics(self) -> _metrics.Metric:
        return self._metrics_type(
            dataset=self._delegate,
            unprivileged_groups=self.unprivileged_groups,
            privileged_groups=self.privileged_groups
        )

    def _attributes_to_propagate_when_wrapping(self):
        return ['unprivileged_groups', 'privileged_groups']


class DatasetWithRegressionMetrics(DatasetWithMetrics):
    def __init__(self, dataset, unprivileged_groups, privileged_groups):
        super().__init__(dataset, _metrics.RegressionDatasetMetric)
        self.unprivileged_groups = unprivileged_groups
        self.privileged_groups = privileged_groups

    def _new_metrics(self) -> _metrics.Metric:
        return self._metrics_type(
            dataset=self._delegate,
            unprivileged_groups=self.unprivileged_groups,
            privileged_groups=self.privileged_groups
        )

    def _attributes_to_propagate_when_wrapping(self):
        return ['unprivileged_groups', 'privileged_groups']


_DATASET_TYPES = {
    "binary label": BinaryLabelDataset,
    "multi class": MulticlassLabelDataset,
    "binary": BinaryLabelDataset,
    "multiclass": MulticlassLabelDataset,
    "multi": MulticlassLabelDataset,
    "regression": RegressionDataset
}


def create_dataset(dataset_type, **kwargs):
    if dataset_type not in _DATASET_TYPES:
        raise ValueError(f"Unknown dataset type: {dataset_type}")
    else:
        if dataset_type == "regression":
            return create_regression_dataset(**kwargs)
        else:
            return create_dataset1(dataset_type, **kwargs)


def create_regression_dataset(unprivileged_groups,
                              privileged_groups,
                              imputation_strategy,
                              **kwargs):
    if kwargs['df'] is None:
        raise ValueError("parameter 'df' must not be None")
    if 'wrap' in kwargs:
        dataset = kwargs.pop('wrap')
    else:
        imputed_df = imputation_strategy(kwargs["df"])
        kwargs['df'] = imputed_df
        protected_attribute_names = []
        privileged_classes = []
        for pg in privileged_groups:
            for key, value in pg.items():
                protected_attribute_names.append(key)
                privileged_classes.append([value])
        kwargs["protected_attribute_names"] = protected_attribute_names
        kwargs["privileged_classes"] = privileged_classes
        dataset = RegressionDataset(**kwargs)
    return DatasetWithRegressionMetrics(dataset,
                                        unprivileged_groups=unprivileged_groups,
                                        privileged_groups=privileged_groups)


def create_dataset1(dataset_type,
                    unprivileged_groups=[{'prot_attr': 0}],
                    privileged_groups=[{'prot_attr': 1}],
                    imputation_strategy=_imputations.DoNothingImputationStrategy,
                    favorable_label=1,
                    unfavorable_label=0,
                    df=None,
                    label_names=['label'],
                    scores_names=None,
                    **kwargs
                    ):
    # imputation_strategy = kwargs.pop('imputation_strategy', _imputations.DoNothingImputationStrategy())
    if 'wrap' in kwargs:
        dataset = kwargs.pop('wrap')
    else:
        imputed_df = imputation_strategy(df)
        kwargs['df'] = imputed_df
        kwargs['favorable_label'] = favorable_label
        kwargs['unfavorable_label'] = unfavorable_label
        kwargs['label_names'] = label_names
        protected_attribute_names = []
        for up in unprivileged_groups:
            for key, _ in up.items():
                protected_attribute_names.append(key)
        kwargs['protected_attribute_names'] = list(set(protected_attribute_names))
        kwargs['scores_names'] = scores_names

        dataset = _DATASET_TYPES[dataset_type](**kwargs)

    binary_label_metrics = [k for k, v in _DATASET_TYPES.items() if
                            v == BinaryLabelDataset or v == MulticlassLabelDataset]
    if dataset_type in binary_label_metrics:
        return DatasetWithBinaryLabelMetrics(dataset, unprivileged_groups, privileged_groups)
    elif dataset_type == "regression":
        return DatasetWithRegressionMetrics(dataset, unprivileged_groups, privileged_groups)


def create_metric(metric, **kwargs):
    metric_type = {
        "regression": _metrics.RegressionDatasetMetric,
        "sample distortion": _metrics.SampleDistortionMetric,
        "binary": _metrics.BinaryLabelDatasetMetric,
        "classification": _metrics.ClassificationMetric,
        "mdss": _metrics.MDSSClassificationMetric,
    }

    ds_1 = kwargs.pop('dataset1')
    if metric_type[metric] in [_metrics.ClassificationMetric,
                               _metrics.SampleDistortionMetric,
                               _metrics.MDSSClassificationMetric]:
        ds_2 = kwargs.pop('dataset2')
        return metric_type[metric](ds_1.delegate, ds_2.delegate, **kwargs)
    else:
        return metric_type[metric](ds_1.delegate, **kwargs)


__all__ = [
    'Dataset',
    'StructuredDataset',
    'RegressionDataset',
    'BinaryLabelDataset',
    'StandardDataset',
    'MulticlassLabelDataset',
    'DatasetWithMetrics',
    'DatasetWithBinaryLabelMetrics',
    'create_dataset',
]

# keep this line at the bottom of this file
aequitas.logger.debug("Module %s correctly loaded", __name__)



import logging
from typing import Any

# import os
# os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'

# keep this line at the top of this file
__all__ = ["logger", "isinstance", "Decorator"]


logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger('aequitas')
"""General logger to be used in all `aequitas*` modules"""


__py_isinstance = isinstance


def isinstance(obj, cls):
    """A version of `isinstance` that takes type unions into account"""

    if hasattr(cls, '__args__') and __py_isinstance(cls.__args__, tuple):
        return any(__py_isinstance(obj, t) for t in cls.__args__)
    return __py_isinstance(obj, cls)


class Decorator:
    # @classmethod
    # def _make_decorator_instance_of_delegate_type(cls, decorated):
    #     delegate = decorated._delegate
    #     delegate_type = type(delegate)
    #     old_is_instance_check = delegate_type.__instancecheck__
    #     if hasattr(old_is_instance_check, '__decorated') and getattr(old_is_instance_check, '__decorated'):
    #         return
    #     def new_instance_check(instance, klass):
    #         if old_is_instance_check(instance, klass):
    #             return True
    #         if isinstance(instance, cls):
    #             return old_is_instance_check(instance._delegate, klass)
    #         return False
    #
    #     setattr(new_instance_check, '__decorated', True)
    #     try:
    #         delegate_type.__instancecheck__ = new_instance_check
    #     except TypeError:
    #         import warnings
    #         warnings.warn(f"Cannot set __instancecheck__ on built-in type {delegate_type.__name__}")

    def __init__(self, delegate) -> None:
        if delegate is None:
            raise ValueError("Delegate cannot be None in decorator pattern")
        self._delegate = delegate
        # self._make_decorator_instance_of_delegate_type(self)

    def __getattr__(self, name):
        return getattr(self._delegate, name)
    
    def __setattr__(self, name, value):
        if name == '_delegate':
            super().__setattr__(name, value)
        elif hasattr(self._delegate, name):
            setattr(self._delegate, name, value)
        else:
            super().__setattr__(name, value)
    
    def __delattr__(self, name):
        if hasattr(self._delegate, name):
            delattr(self._delegate, name)
        else:
            super().__delattr__(name)

    def __getitem__(self, key):
        return self._delegate[key]
    
    def __setitem__(self, key, value):
        self._delegate[key] = value

    def __delitem__(self, key):
        del self._delegate[key]

    def __dir__(self):
        return dir(self._delegate) + super().__dir__()
    
    def __call__(self, *args: Any, **kwds: Any) -> Any:
        return self._delegate(*args, **kwds)
    
    def __len__(self) -> int:
        return len(self._delegate)
    
    def __iter__(self):
        return iter(self._delegate)

    def __eq__(self, value: object) -> bool:
        return self._delegate == value
    
    def __hash__(self) -> int:
        return hash(self._delegate)
    
    def __ne__(self, value: object) -> bool:
        return self._delegate != value
    
    def __lt__(self, value: object) -> bool:
        return self._delegate < value
    
    def __le__(self, value: object) -> bool:
        return self._delegate <= value
    
    def __gt__(self, value: object) -> bool:
        return self._delegate > value
    
    def __ge__(self, value: object) -> bool:
        return self._delegate >= value
    
    def __add__(self, value: object) -> Any:
        return self._delegate + value
    
    def __sub__(self, value: object) -> Any:
        return self._delegate - value
    
    def __mul__(self, value: object) -> Any:
        return self._delegate * value
    
    def __truediv__(self, value: object) -> Any:
        return self._delegate / value
    
    def __floordiv__(self, value: object) -> Any:
        return self._delegate // value
    
    def __mod__(self, value: object) -> Any:
        return self._delegate % value
    
    def __pow__(self, value: object) -> Any:
        return self._delegate ** value
    
    def __lshift__(self, value: object) -> Any:
        return self._delegate << value
    
    def __rshift__(self, value: object) -> Any:
        return self._delegate >> value
    
    def __and__(self, value: object) -> Any:
        return self._delegate & value
    
    def __xor__(self, value: object) -> Any:
        return self._delegate ^ value
    
    def __or__(self, value: object) -> Any:
        return self._delegate | value
    
    def __repr__(self) -> str:
        return repr(self._delegate)
    
    def __str__(self) -> str:
        return str(self._delegate)

    def __instancecheck__(self, instance):
        return super().__instancecheck__(instance)

# keep this line at the bottom of this file
logger.debug("Module %s correctly loaded", __name__)



import aequitas
from aequitas.core.algorithms import AequitasTransformer

from aif360.algorithms.postprocessing import (
    DeterministicReranking as _Aif360DeterministicReranking,
    EqOddsPostprocessing as _Aif360EqOddsPostprocessing,
    RejectOptionClassification as _Aif360RejectOptionClassification,
    CalibratedEqOddsPostprocessing as _Aif360CalibratedEqOddsPostprocessing,
)


class DeterministicReranking(AequitasTransformer):
    def __init__(self, **kwargs):
        super().__init__(_Aif360DeterministicReranking, **kwargs)


class EqOddsPostprocessing(AequitasTransformer):
    def __init__(self, **kwargs):
        super().__init__(_Aif360EqOddsPostprocessing, **kwargs)


class RejectOptionClassification(AequitasTransformer):
    def __init__(self, **kwargs):
        super().__init__(_Aif360RejectOptionClassification, **kwargs)


class CalibratedEqOddsPostprocessing(AequitasTransformer):
    def __init__(self, **kwargs):
        super().__init__(_Aif360CalibratedEqOddsPostprocessing, **kwargs)


# keep this line at the bottom of this file
aequitas.logger.debug("Module %s correctly loaded", __name__)



import aequitas

# noinspection PyUnresolvedReferences
from aif360.metrics import (
    Metric,
    DatasetMetric,
    RegressionDatasetMetric,
    SampleDistortionMetric,
    BinaryLabelDatasetMetric,
    ClassificationMetric,
    MDSSClassificationMetric,
)


# keep this line at the bottom of this file
aequitas.logger.debug("Module %s correctly loaded", __name__)
